from .autoname import AutoNameByCount
from enum import auto

class Command(AutoNameByCount):
    aimode = auto()
    addtime = auto()
    addbot = auto()
    allvice = auto()
    ban = auto()
    cleartempbans = auto()
    cure = auto()
    changeteam = auto()
    changemap = auto()
    confetti = auto()
    clearchat = auto()
    clearrcon = auto()
    currentwave = auto()
    cmdlist = auto()
    comictext = auto()
    copylog = auto()
    drunk = auto()
    deletesave = auto()
    displayinfo = auto()
    decalclear = auto()
    detonatebomb = auto()
    endmatch = auto()
    eventtext = auto()
    echo = auto()
    explode = auto()
    explodebig = auto()
    forceweap = auto()
    flash = auto()
    freeze = auto()
    globalsound = auto()
    grappleplayers = auto()
    givevice = auto()
    kickbots = auto()
    kick = auto()
    kill = auto()
    killenemies = auto()
    killall = auto()
    killteam = auto()
    listdir = auto()
    localsound = auto()
    listsfx = auto()
    listmutators = auto()
    lobby = auto()
    listreckon = auto()
    move = auto()
    money = auto()
    matchtime = auto()
    memoryuse = auto()
    mutatorsetting = auto()
    moveteam = auto()
    moveenemies = auto()
    moveall = auto()
    networkinfo = auto()
    netcodetimerange = auto()
    netcodemargin = auto()
    netcodedamageallow = auto()
    netcodedistance = auto()
    netcodereset = auto()
    netcodedeathpredicttime = auto()
    netcodedeathpredictavg = auto()
    netcodedeathpredictping = auto()
    netcodereckon = auto()
    objreport = auto()
    pm = auto()
    powerup = auto()
    pause = auto()
    poison = auto()
    password = auto()
    playersound = auto()
    rawsay = auto()
    randbkgd = auto()
    resetvicesall = auto()
    resetvicesplayer = auto()
    restartmap = auto()
    restartround = auto()
    rcon = auto()
    rconget = auto()
    revive = auto()
    reviveall = auto()
    resetworld = auto()
    resetplayer = auto()
    resetallplayers = auto()
    resetmutatorsplayer = auto()
    reloadmutators = auto()
    say = auto()
    showcollision = auto()
    showtiles = auto()
    showpath = auto()
    setwave = auto()
    setvice = auto()
    spawnenemy = auto()
    setlife = auto()
    shock = auto()
    showlogic = auto()
    showplayerinfo = auto()
    skip = auto()
    showhitbox = auto()
    stopcmdlist = auto()
    spawnbody = auto()
    shootingrange = auto()
    showdps = auto()
    setrangetargets = auto()
    setmutatorplayer = auto()
    setbotaimode = auto()
    spawnvice = auto()
    showhud = auto()
    setbotaidivert = auto()
    test = auto()
    tempban = auto()
    thaw = auto()
    unbanlast = auto()
    unpause = auto()
    uptime = auto()
    zombending = auto()


from enum import Enum, auto

from .autoname import AutoName


class Hat(AutoName):
    no_hat = auto()
    ballin_cap = auto()
    military_cover = auto()
    constructing_helmet = auto()
    solid_bandana = auto()
    canadian_mask = auto()
    spas_hat = auto()
    the_tipper = auto()
    ten_gallon = auto()
    kings_adornment = auto()
    blue_soldier_aviator = auto()
    goldilocks = auto()
    raggy_rag = auto()
    protector = auto()
    suffocator = auto()
    the_entertainer = auto()
    mrowr_ears = auto()
    samurai_sunhat = auto()
    facial_cover = auto()
    bloody_mess = auto()
    twirly_bean = auto()
    bro_glasses = auto()
    chef_hat = auto()
    effed_up = auto()
    horny_helm = auto()
    cooking_pot = auto()
    party_hat = auto()
    fez = auto()
    night_vision = auto()
    riot_helmet = auto()
    wielders_delight = auto()
    sunny_hat = auto()
    funny_glasses = auto()
    the_brunette = auto()
    the_boss = auto()
    leather_hat = auto()
    sneaky_beaky = auto()
    roberts_hood = auto()
    oooshanka = auto()
    beer_hat = auto()
    duck_mask = auto()
    burger_boy = auto()
    sombreado = auto()
    the_sellout = auto()
    cool_backwards_hat = auto()
    ass_clown_wig = auto()
    paper_boat = auto()
    the_cosmonaut = auto()
    virtualized_goggles = auto()
    medical_mask = auto()
    tinfoil_hat = auto()
    dashing_topper = auto()
    conehead = auto()
    stupid_magic_hat = auto()
    bankroll_visor = auto()
    the_ol_arrow_gag = auto()
    santa_hat = auto()
    elf_ears = auto()
    frostman_mask = auto()
    stag_antlers = auto()
    hanukkah_headband = auto()
    kwanzaa_krown = auto()


import json

from .rcon_model import RconRequest, RconEvent, event_types, hat_types
from .rcon_model import Hat, Weapon, Enemy

from loguru import logger

from .rcon_model.request_cases import RequestCase

def get_coercions():
    return {
        ('Level', 'AttackerID', 'VictimID', 'KillerID', 'PlayerID', 'Bots', 'Alive', 
            'Dead', 'Amount', 'LandingX', 'LandingY', 'Cost', 'Players1',
            'Players2', 'Alive1', 'Alive2', 'Score1', 'Score2') : lambda x : int(x),
        ('X', 'Y', 'FlagX', 'FlagY'): lambda x: float(x),
        
        ('Hat'): lambda x : Hat(int(x)),
        
        ('Weap1', 'Weap2', 'Equip', 'OffWeap', 'OffWeap2', 'KillerWeapon', 'NewWeapon',
            'SkinWeapon',  'Weapon'):  lambda x: Weapon(int(x)),
        
        ('EnemyType',): lambda x : Enemy(x),
        
        ('IsAdmin', 'Teamkill', 'Kicked', 'Drone', 'Autobalanced', 'Headshot', 'Flawless',
            'WasHome', 'Thrown', 'Host'): lambda x: bool(int(x)),
        ('EventID'): lambda x: RconEvent(int(x))
    }


def convert_request_data(request_data: dict):
    request_case = RequestCase(int(request_data['CaseID']))
    if request_case == RequestCase.request_match:
        return event_types.request_data_match(**request_data)
    elif request_case == RequestCase.request_player:
        return event_types.request_data_player(**request_data)
    else:
        return event_types.BaseClass(**request_data)


def is_request_data(message: dict):
    return 'CaseID' in message and 'RequestID' in message


def initialize_class(message: dict):
    """
        Main function: coerces the types of the incoming message to be the types that the classes expect.
    """
    for k, v in message.items():
        if k.endswith('Profile') and 'ProfileID' in v and 'StoreID' in v:
            message[k] = event_types.PlayerProfile(**json.loads(v))
    
    # Convert the various strings to their types
    for m_key, m_f in get_coercions().items():
        for key, val in message.items():
            if key in m_key:
                message[key] = m_f(val)
    
    # Return the request data type instead of normal type
    if is_request_data(message):
        return convert_request_data(message)

    class_name = message['EventID'].name
    module = __import__('bmio.rcon_model.event_types')
    try:
        class_ = getattr(module, class_name)
    except:
        logger.error('Class is not implemented yet')
        return event_types.BaseClass(**message)
    return class_(**message)





from dataclasses import dataclass, asdict
from typing import Optional

from .rcon_events import RconEvent
from .enemy_types import Enemy, EnemyRank
from .weapon_types import Weapon
from .hat_types import Hat


@dataclass
class BaseClass:
    Time: str
    EventID: RconEvent

    def asdict(self):
        return {k: str(v) for k, v in asdict(self).items()}


@dataclass
class PlayerProfile:
    ProfileID: str
    StoreID: str

@dataclass
class server_shutdown(BaseClass):
    """Triggers when the server is shutdown."""

@dataclass
class server_startup(BaseClass):
    """"Triggers when the server is started up"""

@dataclass
class lobby_connect(BaseClass):
    """Triggers when the server connects to the server list."""


@dataclass
class lobby_disconnect(BaseClass):
    """Triggers when the server loses connection to the server list."""


@dataclass
class player_connect(BaseClass):
    """Triggers when a new player connects to the server."""
    IP: str
    PlayerName: str
    PlayerID: int
    Profile: PlayerProfile
    IsAdmin: str


@dataclass
class player_spawn(BaseClass):
    """Triggers when a player or NPC respawns. The weapons returned are not always what the player picks in the loadout, depending on when they finish selecting their loadout. You should use player_loadout for handling any loadout weapons."""
    PlayerID: int
    Profile: PlayerProfile
    X: float
    Y: float
    Hat: Hat
    Hat2: Hat
    Name: str
    Color: str
    Team: str
    Weap1: Weapon
    Weap2: Weapon
    Equip: Weapon
    OffWeap: Weapon
    OffWeap2: Weapon
    EnemyType: Optional[Enemy] = None
    EnemyRank: Optional[EnemyRank] = None


@dataclass
class player_death(BaseClass):	
    """Triggers when a player dies."""
    VictimID: str
    KillerID: str
    AssisterID: str
    VictimProfile: PlayerProfile
    AssisterProfile: PlayerProfile
    KillerWeapon: Weapon
    Headshot: bool
    DeathType: str
    Drone: bool
    Teamkill: str
    VictimX: float
    VictimY: float
    KillerX: Optional[float] = None
    KillerY: Optional[float] = None
    KillerProfile: Optional[PlayerProfile] = None
    AssistX: Optional[float] = None
    AssistY: Optional[float] = None


@dataclass
class player_disconnect(BaseClass):
    """Triggers when a player disconnects from the server."""
    IP: str
    PlayerID: int
    Profile: PlayerProfile
    IsAdmin: bool
    Kicked: bool
    KickReason: str


@dataclass
class player_team_change(BaseClass):
    """Triggers when a player changes team."""
    PlayerID: int
    Profile: PlayerProfile
    OldTeam: str
    NewTeam: str
    Autobalanced: bool


@dataclass
class player_level_up(BaseClass):
    """Triggers when a player levels up."""
    PlayerID: int
    Profile: PlayerProfile
    Level: int
    NewWeapon: Weapon
    SkinWeapon: Weapon
    SkinType: str


@dataclass
class player_get_powerup(BaseClass):
    """Triggers when a player gets a power up"""
    PlayerID: int
    Profile: PlayerProfile
    PowerUp: str
    X: float
    Y: float


@dataclass
class player_damage(BaseClass):
    """Triggers when a player takes damage. DISABLED FOR NOW"""
    AttackerID: int
    VictimID: int
    AttackerProfile: PlayerProfile
    VictimProfile: PlayerProfile
    Headshot: bool
    Damage: int


@dataclass
class player_loaded(BaseClass):
    """Triggers when a player is finished loading their map."""
    PlayerID: int
    Profile: PlayerProfile


@dataclass
class tdm_round_start(BaseClass):
    """Triggers when a Team Deathmatch round starts."""
    Alive1: int
    Alive2: int
    Players1: int
    Players2: int


@dataclass
class tdm_round_end(BaseClass):
    """Triggers when a Team Deathmatch round ends."""
    Players1: int
    Players2: int
    Winner: str
    Score1: int
    Score2: int
    RoundEndType: str
    Flawless: bool


@dataclass
class tdm_flag_unlocked(BaseClass):
    """Triggers when the Team Deathmatch flag unlocks for capture."""
    Alive1: int
    Alive2: int
    Players1: int
    Players2: int
    FlagX: float
    FlagY: float


@dataclass
class tdm_switch_sides(BaseClass):
    """Triggers when the server switches team sides in Team Deathmatch."""
    Score1: int
    Score2: int


@dataclass
class ctf_taken(BaseClass):
    """Triggers when a flag in CTF is stolen."""
    CarrierID: int
    CarrierProfile: PlayerProfile
    FlagTeam: str
    WasHome: bool
    FlagX: float
    FlagY: float


@dataclass
class ctf_dropped(BaseClass):
    """Triggers when a flag in CTF is dropped by a flag carrier."""
    CarrierID: int
    CarrierProfile: PlayerProfile
    FlagTeam: str
    Thrown: bool
    FlagX: float
    FlagY: float


@dataclass
class ctf_returned(BaseClass):
    """Triggers when a flag is returned to home base."""
    ReturnPlayerID: int
    ReturnProfile: PlayerProfile
    ReturnType: str
    FlagX: float
    FlagY: float


@dataclass
class ctf_scored(BaseClass):
    """Triggers when a team scores a point in CTF."""
    CarrierID: int
    CarrierProfile: PlayerProfile
    ScoringTeam: str
    Score1: str
    Score2: str


@dataclass
class ctf_generator_repaired(BaseClass):
    """Triggers when a generator is repaired."""
    ID: str
    Team: str
    RepairerID: str
    RepairerProfile: PlayerProfile


@dataclass
class ctf_generator_destroyed(BaseClass):
    """Triggers when a generator is destroyed."""
    ID: str
    Team: str
    KillerID: str
    KillerProfile: PlayerProfile


@dataclass
class ctf_turret_repaired(BaseClass):
    """Triggers when a turret is repaired."""
    ID: str
    Team: str
    RepairerID: str
    RepairerProfile: PlayerProfile


@dataclass
class ctf_turret_destroyed(BaseClass):
    """Triggers when a turret is destroyed."""
    ID: str
    Team: str
    KillerID: str
    KillerProfile: PlayerProfile


@dataclass
class ctf_resupply_repaired(BaseClass):
    """Triggers when a resupply station is repaired."""
    ID: str
    Team: str
    RepairerID: str
    RepairerProfile: PlayerProfile


@dataclass
class ctf_resupply_destroyed(BaseClass):
    """Triggers when a resupply station is destroyed."""
    ID: str
    Team: str
    KillerID: str
    KillerProfile: PlayerProfile


@dataclass
class match_end(BaseClass):
    """Triggers when the current match ends."""
    WinnerText: str
    WinnerColor: str
    WinnerID: str
    WinnerTeam: str
    GameModeID: str
    NextMapFile: str
    NextMap: str
    PlayerData: str


@dataclass
class match_overtime(BaseClass):
    """Triggers when the match enters over time."""



@dataclass
class match_start(BaseClass):
    """Triggers when a new match starts, does not trigger until warm up phase ends (if there is one)."""
    MapName: str
    MapFile: str
    GameModeID: str
    WorkshopID: str
    MD5: str


@dataclass
class survival_new_wave(BaseClass):
    """Triggers at the start of a new wave in Survival mode."""
    WaveNumber: str
    Enemies: str
    Chests: str
    ChestPrice: str
    ChestCrash: str


@dataclass
class survival_wave_begins(BaseClass):
    """Triggers when the control point flag unlocks for enemies to capture. Turns out this still triggers during Survival Classic or when the prep time is set to 0, will probably changed to save bandwidth."""
    WaveObjective: str
    WaveNumber: str


@dataclass
class survival_buy_chest(BaseClass):
    """Triggers when a player opens a chest."""
    PlayerID: int
    Profile: PlayerProfile
    ChestID: str
    ChestCost: int
    PlayerMoney: int


@dataclass
class log_message(BaseClass):
    """Triggers when a message is logged into the server console."""
    Message: str
    Color: str

@dataclass
class RequestDataBase(BaseClass):
    CaseID: str
    RequestID: str


@dataclass
class request_data_match(RequestDataBase):
    """Triggered when an RCON client makes a request. See below for more documentation on this RCON event."""
    GamemodeName: str
    Team1Score: int
    MaxPlayers: int 
    MaxScore: int
    TimeLeft: int
    TimeStr: str
    Version: str
    Map: str
    MaxTime: int
    ServerName: int
    Team2Score: int
    Players: int
    Overtime: int
    # TODO: make a gamemode enum
    GamemodeID: int
    

@dataclass
class request_data_player(RequestDataBase):
    """Triggered when an RCON client makes a request. See below for more documentation on this RCON event."""
    ID: str
    Name: str
    Color: str
    Team: str
    Kills: int
    Deaths: int
    Assists: int
    Score: int
    Profile: str
    Store: str
    Alive: str
    Bot: str
    Hat: Hat
    Hat2: Hat
    Money: int
    RespawnCost: int
    Premium: str
    
    GamemodeID: str
    X: Optional[int] = None
    Y: Optional[int] = None
    WeaponsDealRank: Optional[str] = None
    ClanID: Optional[str] = None
    ClanTag: Optional[str] = None

    


@dataclass
class command_entered(BaseClass):
    """Triggered when a command is entered into the console."""
    Command: str
    Source: str
    ReturnText: Optional[str] = None


@dataclass
class rcon_logged_in(BaseClass):
    """Triggers when an RCON client successfully logs in."""
    RconIP: str
    RconPort: str
    RconSocket: str
    GameModeID: str
    MapName: str


@dataclass
class match_paused(BaseClass):
    """Triggered when the server is paused."""


@dataclass
class match_unpaused(BaseClass):
    """Triggered when the server is unpaused."""


@dataclass
class warmup_start(BaseClass):
    """Triggered when the warm up phase has begun."""
    WarmupTime: str


@dataclass
class rcon_disconnect(BaseClass):
    """Triggers when an RCON client disconencts."""
    RconIP: str
    RconPort: str
    RconSocket: str


@dataclass
class rcon_ping(BaseClass):
    """Triggers every 5 seconds for each connected RCON client."""


@dataclass
class chat_message(BaseClass):
    """Triggers when a player sends a chat message in the Server tab."""
    PlayerID: int
    Name: str
    Profile: PlayerProfile
    Message: str
    Team: str
    NameColor: str


@dataclass
class survival_get_vice(BaseClass):
    """Triggers when a player collects a vice in Survival mode."""
    PlayerID: int
    Profile: PlayerProfile
    ViceID: str
    Amount: int
    X: float
    Y: float


@dataclass
class survival_use_vice(BaseClass):
    """Triggers when a player uses a consumable vice. (Rubbing Alcohol, Smokes, Hot Wings, etc)"""
    PlayerID: int
    Profile: PlayerProfile
    ViceID: str


@dataclass
class survival_player_revive(BaseClass):
    """Triggers when a player is revived outside of a new wave starting."""
    RevivingPlayerID: int
    SaviorPlayerID: int
    RevivingProfile: PlayerProfile
    SaviorProfile: PlayerProfile
    Antacids: str
    Cost: int


@dataclass
class player_taunt(BaseClass):
    """Triggers when a player uses a emote, such as /drink or /smoke"""
    PlayerID: int
    Profile: PlayerProfile
    TauntID: str


@dataclass
class survival_complete_mission(BaseClass):
    """Triggered when a player completes a mission from the bar in Survival."""
    PlayerID: int
    Profile: PlayerProfile
    Amount: int
    Vice: str
    Type: str


@dataclass
class survival_take_mission(BaseClass):
    """Triggered when a player accepts a mission from the bar in Survival."""
    PlayerID: int
    Profile: PlayerProfile
    Amount: int
    Vice: str
    Type: str


@dataclass
class survival_fail_mission(BaseClass):
    """Triggered when a player fails or abandons a mission from the bar in Survival."""
    PlayerID: int
    Profile: PlayerProfile
    Amount: int
    Vice: str
    Type: str


@dataclass
class zombrains_revive(BaseClass):
    """"""

@dataclass
class Triggered(BaseClass):
    """zombie player revives from killing players (or through console command)"""
    PlayerID: int
    Profile: PlayerProfile


@dataclass
class zombrains_buy_weapon(BaseClass):
    """Triggered when a player buys a weapon from a weapon printer."""
    PlayerID: int
    Profile: PlayerProfile
    Weapon: Weapon
    Cost: int


@dataclass
class zombrains_begin(BaseClass):
    """Triggered when the match starts in Zombrains."""


@dataclass
class zombrains_helicopter_arriving(BaseClass):
    """Triggered when the helicopter spawns to pick up the surviving humans in Zombrains."""
    LandingX: float
    LandingY: float


@dataclass
class zombrains_helicopter_boarding(BaseClass):
    """Triggered when the helicopter reaches the landing zone and is accepting human players to board in Zombrains/
    The helicopter doesn't always stop exactly at the coordinates reported in the zombrains_helicopter_arriving event, 
    so thats why its position is recorded here as well."""
    X: float
    Y: float


@dataclass
class zombrains_helicopter_player_boarded(BaseClass):
    """Triggered when a human player boards the helicopter in Zombrains."""
    PlayerID: int
    Profile: PlayerProfile


@dataclass
class zombrains_end(BaseClass):
    """Triggered when the Zombrains match ends.
    The zombies win if "Alive" returns as "0", the humans win if "Alive" is anything higher then "0"."""
    Alive: int
    Dead: int


@dataclass
class game_over(BaseClass):
    """Triggered the very same frame the game has ended. More precise then the match_end RCON event, but does not include PlayerData."""
    WinnerText: str
    WinnerColor: str
    WinnerID: int
    WinnerTeam: str
    GameModeID: str
    CurrentMapName: str
    CurrentMapFile: str


@dataclass
class server_empty(BaseClass):
    """Triggered when the last human player (except the host player) has quit. You can check if the server is literally empty by checking that the 'Bots' and 'Host' key both return '0'."""
    Bots: int
    Host: bool


@dataclass
class weaponsdeal_rankchange(BaseClass):
    """Triggered when a player ranks up or down in Weapons Deal and gets a new weapon. I guess I should add the weapon ID too at some point."""
    PlayerID: int
    Profile: PlayerProfile
    WeaponsDealRank: str


@dataclass
class takeover_flagcapture(BaseClass):
    """Triggered when a team caps a flag in Take Over."""
    FlagID: str
    FlagX: float
    FlagY: float
    NewOwner: str
    LastOwner: str
    FlagsTeamOne: str
    FlagsTeamTwo: str


@dataclass
class takeover_flagscreated(BaseClass):
    """Triggered when the match starts in Take Over, or when the flags are randomly cycled."""
    FlagAmount: int
    Team1Score: str
    Team2Score: str
    FlagData: str
    FlagID: str
    FlagX: float
    FlagY: float


@dataclass
class player_loadout(BaseClass):
    """Triggered when a player finishes using the load out menu. This is similar to player_spawn, but sometimes players do not select their loadout until after they spawn."""
    PlayerID: int
    Profile: PlayerProfile
    Weap1: Weapon
    Weap2: Weapon
    Dualwield: str
    Equip: str
    OffWeap: str
    OffWeap2: Weapon


@dataclass
class survival_bomb_defused(BaseClass):
    """Triggered when a bomb is defused in Survival."""
    TimeLeft: str
    Defuser: str


@dataclass
class survival_bomb_exploded(BaseClass):
    """Triggered when a bomb explodes in Survival and ends the match. This event will trigger before any of the match ending RCON events are triggered."""


@dataclass
class survival_bomb_rearmed(BaseClass):
    """Triggered when Bomb Dude, Demolitions Guy, Operator or EXPLODEBOT 5000 re-arm a defused bomb."""
    TimeLeft: str


@dataclass
class sleep(BaseClass): 
    """the dedicated server falls sleep to save power."""




from enum import Enum

class RconEvent(Enum):
    server_startup = 0
    server_shutdown = 1
    lobby_connect = 2
    lobby_disconnect = 3
    player_connect = 4
    player_spawn = 5
    player_death = 6
    player_disconnect = 7
    player_team_change = 8
    player_level_up = 9
    player_get_powerup = 10
    player_damage = 11
    player_loaded = 12
    tdm_round_start = 13
    tdm_round_end = 14
    tdm_flag_unlocked = 15
    tdm_switch_sides = 16
    ctf_taken = 17
    ctf_dropped = 18
    ctf_returned = 19
    ctf_scored = 20
    ctf_generator_repaired = 21
    ctf_generator_destroyed = 22
    ctf_turret_repaired = 23
    ctf_turret_destroyed = 24
    ctf_resupply_repaired = 25
    ctf_resupply_destroyed = 26
    match_end = 27
    match_overtime = 28
    match_start = 29
    survival_new_wave = 30
    survival_flag_unlocked = 31
    survival_buy_chest = 32
    log_message = 33
    request_data = 34
    command_entered = 35
    rcon_logged_in = 36
    match_paused = 37
    match_unpaused = 38
    warmup_start = 39
    rcon_disconnect = 40
    rcon_ping = 41
    chat_message = 42
    survival_get_vice = 43
    survival_use_vice = 44
    survival_player_revive = 45
    player_taunt = 46
    survival_complete_mission = 47
    survival_take_mission = 48
    survival_fail_mission = 49
    zombrains_revive = 50
    zombrains_buy_weapon = 51
    zombrains_begin = 52
    zombrains_helicopter_arriving = 53
    zombrains_helicopter_boarding = 54
    zombrains_helicopter_player_boarded = 55
    zombrains_end = 56
    game_over = 57
    server_empty = 58
    weaponsdeal_rankchange = 59
    takeover_flagcapture = 60
    takeover_flagscreated = 61
    player_loadout = 62
    survival_bomb_defused = 63
    survival_bomb_exploded = 64
    survival_bomb_rearmed = 65
    sleep = 66
    wake = 67



from .rcon_events import RconEvent
from .rcon_requests import RconRequest
from .hat_types import Hat
from .weapon_types import Weapon
from .enemy_types import Enemy, EnemyRank
from .event_types import *
from .command_types import Command

__all__ = [
    'RconEvent',
    'RconRequest',
    'Hat',
    'Enemy', 
    'EnemyRank', 
    'BaseClass',
    'Command',
    'PlayerProfile',
    'server_shutdown',
    'lobby_connect',
    'lobby_disconnect',
    'player_connect',
    'player_spawn',
    'player_death',
    'player_disconnect',
    'player_team_change',
    'player_level_up',
    'player_get_powerup',
    'player_damage',
    'player_loaded',
    'tdm_round_start',
    'tdm_round_end',
    'tdm_flag_unlocked',
    'tdm_switch_sides',
    'ctf_taken',
    'ctf_dropped',
    'ctf_returned',
    'ctf_scored',
    'ctf_generator_repaired',
    'ctf_generator_destroyed',
    'ctf_turret_repaired',
    'ctf_turret_destroyed',
    'ctf_resupply_repaired',
    'ctf_resupply_destroyed',
    'match_end',
    'match_overtime',
    'match_start',
    'survival_new_wave',
    'survival_wave_begins',
    'survival_buy_chest',
    'log_message',
    'request_data',
    'command_entered',
    'rcon_logged_in',
    'match_paused',
    'match_unpaused',
    'warmup_start',
    'rcon_disconnect',
    'rcon_ping',
    'chat_message',
    'survival_get_vice',
    'survival_use_vice',
    'survival_player_revive',
    'player_taunt',
    'survival_complete_mission',
    'survival_take_mission',
    'survival_fail_mission',
    'zombrains_revive',
    'Triggered',
    'zombrains_buy_weapon',
    'zombrains_begin',
    'zombrains_helicopter_arriving',
    'zombrains_helicopter_boarding',
    'zombrains_helicopter_player_boarded',
    'zombrains_end',
    'game_over',
    'server_empty',
    'weaponsdeal_rankchange',
    'takeover_flagcapture',
    'takeover_flagscreated',
    'player_loadout',
    'survival_bomb_defused',
    'survival_bomb_exploded',
    'survival_bomb_rearmed',
    'sleep'
]


from enum import Enum

class Weapon(Enum):
    none = -1
    fists = 0
    tazer = 1
    pistol = 2
    silenced_pistol = 3
    underwater_pistol = 4
    revolver = 5
    burst_pistol = 6
    compact_pistol = 7
    handcannon = 8
    uzi = 9
    compact_uzi = 10
    fire_uzi = 11
    light_smg = 12
    smg = 13
    power_smg = 14
    long_smg = 15
    sawn_off = 16
    pump = 17
    db = 18
    trench = 19
    auto_shotgun = 20
    light_ar = 21
    assault_rifle = 22
    keymaster = 23
    heal_rifle = 24
    underwater_rifle = 25
    musket = 26
    power_rifle = 27
    scoped_rifle = 28
    long_rifle = 29
    lever = 30
    plinger = 31
    sniper_rifle = 32
    bolt_action = 33
    tranq_rifle = 34
    lightning_gun = 35
    machine_gun = 36
    chain_gun = 37
    grenade_launcher = 38
    heavy_gl = 39
    gas_launcher = 40
    rocket_launcher = 41
    light_rocket = 42
    mortar = 43
    plasma_rifle = 44
    acid_gun = 45
    goo_gun = 46
    muleslapper = 47
    fusion_cannon = 48
    bow = 49
    crossbow = 50
    grapplehook = 51
    knife = 52
    sword = 53
    wrench = 54
    # nothing = 55
    magic = 56
    railgun = 57
    buzzer = 58
    blaster = 59
    riot_shield = 60
    drone = 61
    frag = 62
    potato_masher = 63
    emp = 64
    molotov = 65
    nitrogen_mine = 66
    healnade = 67
    gasnade = 68
    flashbang = 69
    skate = 70
    tomahawk = 71
    suicide_vest = 72
    muleslapper_sentry = 73
    pineapple = 74
    scuba = 75
    dynamite = 76
    kunai = 77
    jetpack = 78
    boomerang = 79
    unknown = 80




from enum import Enum


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return count


class AutoNameByCount(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


from enum import Enum, auto

from .autoname import AutoName


class Enemy(AutoName):
    cannibal = auto(),
    hopper = auto(),
    blue_soldier = auto(),
    purple = auto(),
    sniper = auto(),
    flesheater = auto(),
    blue_lieutenant = auto(),
    fuschia = auto(),
    leaper = auto(),
    bomb_dude = auto(),
    demolition_guy = auto(),
    ninja = auto(),
    samurai = auto(),
    indigo = auto(),
    blue_captain = auto(),
    grandmaster = auto(),
    explodebot_5000 = auto(),
    anthropophagite = auto(),
    moxxy = auto(),
    roxxy = auto(),
    disciple = auto(),
    manling = auto(),
    operator = auto(),
    cowboy = auto(),
    archer = auto(),
    zombie = auto(),
    zhost = auto(),
    zpitter = auto(),
    zomikaze = auto(),
    doctor = auto(),

class EnemyRank(AutoName):
    normal = auto(),
    strong = auto(),
    elite = auto(),
    powerful = auto(),
    god_like = auto()


from queue import Queue
import socket
import struct
import random
import string
from threading import Thread
import json
from typing import Callable

from .rcon_model import RconRequest, RconEvent, RequestDataBase


from loguru import logger

from .data_coerce import initialize_class
from .rcon_model.command_types import Command


START_DELIMITER = b'\xe2\x94\x90'
END_DELIMITER = b'\xe2\x94\x94'

class Bmio:

    def __init__(
        self,
        host = 'localhost',
        port = 42070,
        password = 'admin',
        parallelism = 2,

    ):
        """An implementation of the Producer Consumer Queue for Rcon Objects."""
        self.sock = self.__connect(host, port, password)
        self.send_request(password, RconRequest.login)
        self.event_handlers = {}
        self.request_handlers = {}
        self.packet_queue = Queue(0)
        self.parallelism = parallelism


    def run(self):
        """Begins the processing of Rcon Packets"""
        writer_thread = Thread(target = self.__threadwrap(self.__start_read), args= (self,), daemon = True)
        writer_thread.start()
        
        for i in range(0, self.parallelism):
            reader_thread = Thread(target = self.__threadwrap(self.__handle_events), args= (self,), daemon = True)
            reader_thread.start()

        writer_thread.join()
        reader_thread.join()


    def handler(self, event: RconEvent):
        """Decorator for registering a handler"""
        def add_handler(event, f):
            """"Register a handler"""
            if not event in self.event_handlers:
                self.event_handlers[event] = [f]
            else:
                self.event_handlers[event].append(f)
        def decorator(f):
            add_handler(event, f)
            return f
        return decorator
    


    def send_command(self, command: Command, *args):
        """
            Sends a command over to the server
            
                Parameters:
                    command: The type of command
                    args*
        """
        full_command = command.value
        for arg in args:
            full_command += f' "{arg}"'
        self.send_request(full_command, RconRequest.command)


    def request_data(
        self, 
        request_type: RconRequest, 
        callback: Callable,
        request_params: str = "None"
    ):
        """
            Request game data from the server. When the data is returned, the callback function is called
            
                Parameters:
                    request_type: The type of request
                    callback: function that will be called when the data is returned
                    request_params: optional arguments that come with the request 
        """
        request_id = self.__generate_hash()
        self.send_request_with_id(
            request_id,
            request_params,
            request_type
        )
        self.request_handlers[request_id] = callback


    def __generate_hash(self):
        return ''.join(random.choices(string.ascii_lowercase, k=5))


    def __connect(self, host: str, port: int, password: str):
        """Get a _connection to the boring man rcon"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))
        return sock


    def __handle_events(self):
        """Handle packet events as they stream in"""
        while self.packet_queue:
            packet = self.packet_queue.get()
            
            if isinstance(packet, RequestDataBase):
                if packet.RequestID in self.request_handlers:
                    f = self.request_handlers.pop(packet.RequestID)
                    f(packet)

            elif packet.EventID in self.event_handlers:
                handlers = self.event_handlers[packet.EventID]
                for f in handlers:
                    f(packet)


    def send_request(self, request_data: str, request_type: RconRequest):
        """Send a simple request"""
        request_message = request_data + "\00" 
        request = struct.Struct(
            f'h{ len(bytes(request_message, "utf-8")) }s'
        ).pack(
            request_type.value, request_message.encode('utf-8')
        ) 
        self.sock.send(request)


    def send_request_with_id(self, request_id: str, request_params: str, request_type: RconRequest):
        """Send a request that comes with a request id
            Parameters
                request_id: The unique id associated with this request. Will be returned alongside the data
                request_params: optional parameters associated with the request
                request_type: the RconRequest enum of the request
        """
        request_message = f'"{request_id}" "{request_params}"'
        self.send_request(request_message, request_type)


    def __start_read(self):
        """Start reading and only stop when the delimiters are not present"""
        buffer = self.sock.recv(1024)
        while buffer.find(END_DELIMITER) != -1 and buffer.find(START_DELIMITER) != -1:

            start_index = buffer.find(START_DELIMITER)
            end_index = buffer.find(END_DELIMITER) + len(END_DELIMITER)

            data = buffer[start_index:end_index]
            buffer = buffer[end_index:]
            if data:
                data_info = struct.unpack_from('<'+'3s'+'h', data, 0)
                event_data = struct.unpack_from(
                    '<'+'3s'+'h'+'h'+str(data_info[1])+'s', data, 0)
                event_id = event_data[2]
                message_string = event_data[3].decode().strip()
                message_string = message_string[:-1]
                js = json.loads(message_string)
                logger.debug(js)
                self.packet_queue.put(initialize_class(js))
                if event_id == RconEvent.rcon_ping.value:
                    self.send_request("None", RconRequest.ping)
            buffer += self.sock.recv(1024)


    def __threadwrap(self, threadfunc):
        """"Wrap threads that should be restarted"""
        def wrapper(self):
            while True:
                try:
                    threadfunc()
                except BaseException as e:
                    logger.error('{!r}; restarting thread'.format(e))
                else:
                    logger.error(f'Thread exited normally: {threadfunc.__name__}')
        return wrapper


    